/*
*
* Copyright (c) 1997 Jan Kr"uger <Jan.Krueger@stud.uni-hannover.de>
*
* In short: you can do whatever you want with this, but don't blame me!
*
* THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
* OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
* LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
* OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGE.
*
*/

#ifdef MAP_STORAGE

#include "sendmail.h"

#ifdef MAP_STORAGE_MD5
#include "global.h"
#include "md5.h"

/*
 uncomment and change the secret */
#define SECRET "change this secret"


#define KEY_DELIM	'+'
#define KEYLEN 12   	/* 1 .. 12 */
#endif /* MAP_STORAGE_MD5 */

#define DEF_EQUALSTR	"@EQUAL"

#define CHAR1	'@'
#define CHAR0	'.'

#define DEF_STORAGE_SIZE	100

struct storage_map {
int storage_size;
char *header_field;
char *equalstr;
char *storage;
};

static struct storage_map **storage_h_fields = NULL;
static int nr_h_fields = 1;

int storage_map_init(map, ap)
	MAP *map;
	char *ap;
{
	char *p = ap;
	char *size_str = NULL;
	char *header_str = NULL;
	char *equal_str = NULL;
	struct storage_map *storage_p;

	storage_p = (struct storage_map *)xalloc(sizeof(struct storage_map));
	storage_p->header_field = NULL;
	storage_p->storage_size = DEF_STORAGE_SIZE;

	for (;;)
        {
		while (isascii(*p) && isspace(*p))
			p++;
		if (*p != '-')
			break;
		switch (*++p)
		{
			case 's': /* size */
				size_str = ++p;
				break;

			case 'h': /* header */
				header_str = ++p;
				break;

			case 'e': /* equal for strcmp */
				equal_str = ++p;
				break;
		}
                while (*p != '\0' && !(isascii(*p) && isspace(*p)))
                        p++;
                if (*p != '\0')
                        *p++ = '\0';
	}
	if(size_str != NULL)
		storage_p->storage_size = atoi(size_str);

	if(equal_str != NULL)
		storage_p->equalstr = newstr(equal_str);
	else
		storage_p->equalstr = DEF_EQUALSTR;

	storage_p->storage = xalloc(storage_p->storage_size + 1);

	if(header_str != NULL)
	{
		storage_p->header_field = newstr(header_str);
		storage_h_fields = 
			(struct storage_map **)realloc(storage_h_fields, 
				++nr_h_fields * sizeof(struct storage_map *));
		if(storage_h_fields == NULL)
		{
			syserr("map_storage: out of memory");
			return FALSE;
		}
		storage_h_fields[nr_h_fields - 2] = storage_p;
		storage_h_fields[nr_h_fields - 1] = NULL;
	}

	snprintf(storage_p->storage, storage_p->storage_size, "%s", p);

	if(tTd(73, 1))
		printf("storage_map_init: len %d value '%s'\n",
			storage_p->storage_size, storage_p->storage);

	map->map_file = (char *)storage_p;
	
	return TRUE;
}


char *storage_rewrite_ip_bin(ip)
	char *ip;
{
	static char buf[32 + 1];
	register char *dp = buf;
	int b[4];
	int byte;
	register int bit;

	if(tTd(73, 1))
		printf("ip_to_bin %s\n", ip);

	if(sscanf(ip, "%d.%d.%d.%d", &b[3], &b[2], &b[1], &b[0]) != 4)
		return ip;
	
	for(byte = 3; byte >= 0; byte--)
	{
		register int rb = b[byte];
		for(bit = 8; bit > 0; bit--)
		{
			if((rb & 0x80) == 0x80)
				*dp++ = CHAR1;
			else	
				*dp++ = CHAR0;

			rb <<= 1;
		}
	}

	*dp = '\0';

	return buf;
}

#ifdef MAP_STORAGE_MD5
#define three2four(n) retbuf[n]   = b64[ sig[n] >> 2 ]; \
	retbuf[n+1] = b64[ ((sig[n] << 4) & 0x3f)   |  (sig[n+1] >> 4) ]; \
	retbuf[n+2] = b64[ ((sig[n+1] << 2) & 0x3f) |  (sig[n+2] >> 6) ]; \
	retbuf[n+3] = b64[ (sig[n+2]  & 0x3f) ];  

char *md5_base64(sig)
	unsigned char *sig;
{
	char *b64 = "abcdefghijklmnopqrstuvwxyz"
		    "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789{}";

	static char retbuf[13];

	three2four(0);
	three2four(4);
	three2four(8);
	retbuf[12] = '\0';

	return retbuf;
}

char *storage_rewrite_md5(key)
	char *key;
{
	static char crypted[13];
	unsigned char digest[16];
	MD5_CTX context;

	MD5Init(&context);
	MD5Update(&context, key, strlen(key));
	MD5Final(digest, &context);

	strcpy(crypted, md5_base64(digest));
	crypted[KEYLEN] = '\0';

	return crypted;
}

char *storage_rewrite_magic_md5(key)
	char *key;
{
	char buf[MAXLINE];
	snprintf(buf, sizeof(buf), "%s%s", key, SECRET);
	return storage_rewrite_md5(buf);
}

int storage_test_magic_md5(key)
	char *key;
{
	char *dl;
	if((dl = strchr(key, KEY_DELIM)) == NULL)
		return FALSE;

	*dl++ = '\0';
	if(strcmp(storage_rewrite_magic_md5(key), dl) == 0)
		return TRUE;

	return FALSE;
}
#endif 

char *storage_map_lookup(map, key, av, statp)
	MAP *map;
	char *key;
	char **av;
	int *statp;
{
	struct storage_map *storage_p;
	bool assignmode = FALSE;

	if(tTd(73, 1))
	{
		char **cpp;
		printf("storage_map_lookup: key '%s'\n", key);
		for(cpp = av; cpp && *cpp; cpp++)
			printf("storage_map_lookup: arg '%s'\n", *cpp);
	}

	storage_p = (struct storage_map *)map->map_file;

	if(av && av[0] && av[1])
	{ /* has option */
		if(key[0] != '\0')
		{ /* assign with key */
			switch(av[1][0])
			{
				case 'b': /* convert ip to "@."-binary */
				snprintf(storage_p->storage, 
					storage_p->storage_size, "%s", 
					storage_rewrite_ip_bin(key));
				break;

#ifdef MAP_STORAGE_MD5
				case '5': /* md5 hash */
				snprintf(storage_p->storage, 
					storage_p->storage_size, "%s", 
					storage_rewrite_md5(key));
				break;

				case 'm': /* magic md5 test */
				if(storage_test_magic_md5(key))
				{
					return map_rewrite(map, 
						storage_p->equalstr, 
						strlen(storage_p->equalstr), 
						NULL);
				}
					return NULL;
				break;

#ifdef MAP_STORAGE_MD5_M
				case 'M': /* magic md5 hash */
				snprintf(storage_p->storage, 
					storage_p->storage_size, "%s", 
					storage_rewrite_magic_md5(key));
				break;
#endif /* MAP_STORAGE_MD5_M */
#endif /* MAP_STORAGE_MD5 */
	
				case 's': /* strcmp */
				if(strcmp(storage_p->storage, key) == 0)
				{
					return map_rewrite(map, 
						storage_p->equalstr, 
						strlen(storage_p->equalstr), 
						NULL);
				}
				else
					return NULL;
				break;

				case 'S': /* strcasecmp */
				if(strcasecmp(storage_p->storage, key) == 0)
				{
					return map_rewrite(map, 
						storage_p->equalstr, 
						strlen(storage_p->equalstr), 
						NULL);
				}
				else
					return NULL;
				break;


			}
		}
		else
		{ /* no key */
			switch(av[1][0])
			{
				case 'c': /* clear storage */
				storage_p->storage[0] = '\0';
				break;
			}
		}
	}
	else
	{ /* no option */
		if(key[0] != '\0')
			snprintf(storage_p->storage, 
				storage_p->storage_size, "%s", key);
	}
	return map_rewrite(map, storage_p->storage, 
			strlen(storage_p->storage), NULL);
}

void map_storage_add_header(e)
ENVELOPE *e;
{
	register struct storage_map **sp;
	if(storage_h_fields == NULL)
		return;
	
	for(sp = storage_h_fields; *sp; sp++)
	{
		char *hfield = (*sp)->header_field;
		char *hvalue = (*sp)->storage;

		if(hvalue[0] != '\0')
		{
			addheader(hfield, hvalue, &e->e_header);

			sm_syslog(LOG_NOTICE, e->e_id,
			   "storage: header added field=%s value=%s",
			   hfield, hvalue);

			hvalue[0] = '\0';
		}
	}
	
}
#endif /* MAP_STORAGE */
